--
--    FILE:     Tectonics.lua
--    AUTHOR:  Laurent Di Cesare
--    PURPOSE: Map script simulating plate tectonics for nice mountain ranges
------------------------------------------------------------------------------
--    Copyright (c) 2010 Laurent Di Cesare All rights reserved.
------------------------------------------------------------------------------
--      HISTORY
--      VERSION: 1.0
------------------------------------------------------------------------------

include("math")
include("MapGenerator")
include("FeatureGenerator")
include("TerrainGenerator")

function GetMapScriptInfo()
    return {
        Name = "Tectonics",
        Description = "Tectonics",
        IconIndex = 5,
    }
end
local Wait = function(arg) -- debugging function to be able to use the *** console
    if arg then print(arg) end
    --[[
    local silly = 0
    while silly < 9900000 do
        silly = silly + 1
    end
    --]]
end
local Voronoi =
{
    plotTypes = {}, -- map generation result
    plate = {}, -- plate of plot
    plateCollision = {}, -- base height increase of each plate if it collides with someone
    heightMap = {},
    collisionHeight = 20, -- height added when 2 plates collide
    seaProportion = 0.75, -- fraction of plates tagged as sea plates
    seaLevel = 0, -- base sea altitude
    landLevel = 10, -- base land altitude
    seaNoise = 3, -- max random sea altitude
    landNoise = 4, -- max random land altitude
    landPlotHeight = 10, -- default land height, shouldn't be used if proportions are used instead
    hillPlotHeight = 15,
    mountainPlotHeight = 22,
    seaProportion = 0.65,
    landProportion = 0.22,
    hillProportion = 0.08,
    Create = function(self)
        self.iW, self.iH = Map.GetGridSize()
        self:GenerateSeeds()
        self:FillPlates()
        self:FillPlateAltitudes()
        self:MovePlates()
        self:Spread()
        self:Spread()
        self:MovePlates()
        self:Hotspots()
        self:ComputePlots()
        return self.plotTypes
    end,
    GenerateSeeds = function(self)
        self.plateNumber = self.iH / 10 + self.iW / 5
        self.seaPlates = math.floor( self.plateNumber * self.seaProportion )
        self.expectedSea = math.floor( self.iH * self.iW * self.seaProportion )
        self.expectedLand = math.floor( self.iH * self.iW * self.landProportion )
        self.expectedHill = math.floor( self.iH * self.iW * self.hillProportion )
        for i = 0, self.seaPlates - 1 do
            local plotIndex = Map.Rand(self.iH, "y for seed") * self.iW + Map.Rand(self.iW, "x for seed")
            while self.plate[plotIndex] do
                plotIndex = Map.Rand(self.iH, "y for seed") * self.iW + Map.Rand(self.iW, "x for seed")
            end
            self.plate[plotIndex] = i
        end
        local landHemisphere = Map.Rand(10, "hemisphere for more land")
        local yLandOffset = -1
        if landHemisphere > 4 then yLandOffset = 1 end
        for i = self.seaPlates, self.plateNumber do
            local hemisphere = Map.Rand(10, "hemisphere for land plate")
            if hemisphere > 7 then yLandOffset = yLandOffset * -1 end
            local yMin = math.floor(self.iH/10)
            local yVar = math.floor(self.iH/2)
            local RandY = function(self)
                local y = yMin + Map.Rand(yVar, "y for seed")
                if yLandOffset > 0 then
                    y = self.iH - y
                end
                return y
            end
            local plotIndex = RandY(self) * self.iW + Map.Rand(self.iW, "x for seed")
            while self.plate[plotIndex] do
                plotIndex = RandY(self) * self.iW + Map.Rand(self.iW, "x for seed")
            end
            self.plate[plotIndex] = i
        end
    end,
    South = function(self, x, y)
        if y == 0 then return nil end
        return (y-1)*self.iW + x + 1
    end,
    North = function(self, x, y)
        if y == self.iH - 1 then return nil end
        return (y+1)*self.iW + x + 1
    end,
    West = function(self, x, y)
        if x == 0 then return y*self.iW + self.iW - 1 +1 end
        return y*self.iW + x - 1 + 1
    end,
    East = function(self, x, y)
        if x == self.iW - 1 then return y*self.iW + 1 end
        return y*self.iW + x + 1 + 1
    end,
    NorthWest = function(self, x, y)
        if y == self.iH - 1 then return nil end
        return self:West(x, y + 1)
    end,
    SouthWest = function(self, x, y)
        if y == 0 then return nil end
        return self:West(x, y - 1)
    end,
    Neighbours = function(self, x, y)
        local result = {}
        result["South"] = self:South(x, y)
        result["North"] = self:North(x, y)
        result["West"] = self:West(x, y)
        result["East"] = self:East(x, y)
        result["NorthWest"] = self:NorthWest(x, y)
        result["SouthWest"] = self:SouthWest(x, y)
        return result
    end,
    FillPlates = function( self )
        local continue = true
        while continue do
            continue = false
            local newPlate = {}
            for y = 0, self.iH - 1 do
                for x = 0, self.iW - 1 do
                    local plotIndex = y * self.iW + x + 1
                    local currentPlate = self.plate[plotIndex]
                    -- testing currentPlate first would work and be slightly faster I believe
                    -- but it'd look harder to understand what I'm doing
                    newPlate[plotIndex] = newPlate[plotIndex] or currentPlate
                    if currentPlate then
                        for _, i in pairs(self:Neighbours(x, y)) do
                            if not self.plate[i] then
                                if Map.Rand(10, "randomness in plate size") > 4 then
                                    newPlate[i] = currentPlate
                                end
                            end
                        end
                    end
                end
            end
            self.plate = newPlate
            for i = 1, self.iH * self.iW do
                if not newPlate[i] then continue = true break end
            end
        end
    end,
    FillPlateAltitudes = function(self)
        for i = 0, self.seaPlates - 1 do
            self.plateCollision[i] = self.seaLevel + Map.Rand(self.seaNoise, "sea level collision")
        end
        for i = self.seaPlates, self.plateNumber do
            self.plateCollision[i] = self.collisionHeight
        end
        for plotIndex = 1, self.iH * self.iW do
            if self.plate[plotIndex] < self.seaPlates then
                self.heightMap[plotIndex] = self.seaLevel + Map.Rand(self.seaNoise, "sea level plate variation")
            else
                self.heightMap[plotIndex] = self.landLevel + Map.Rand(self.landNoise, "land level plate variation")
            end
        end
    end,
    MovePlates = function(self)
        local direction = {}
        for i = 0, self.plateNumber do
            direction[i] = 1 + Map.Rand(5, "direction for plate")
        end
        local directionFunctions = { self.North, self.South, self.East, self.West, self.NorthWest, self.SouthWest }
        Wait( "Moving plates" )
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local currentPlate = self.plate[plotIndex]
                local dirFunction = directionFunctions[direction[currentPlate]]
                local destinationPlot = dirFunction(self, x, y)
                if destinationPlot and (self.plate[destinationPlot] ~= currentPlate) then
                    self.heightMap[plotIndex] = self.heightMap[plotIndex] + self.plateCollision[currentPlate] + self.plateCollision[self.plate[destinationPlot]]
                end
            end
        end
    end,
    Spread = function(self)
        local heightMap = {}
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local height = self.heightMap[plotIndex]
                local neighbours = 1
                for dir, i in pairs(self:Neighbours(x, y)) do
                    height = height + self.heightMap[i]
                    neighbours = neighbours + 1
                end
                heightMap[plotIndex] = math.floor(height/neighbours)
            end
        end
        self.heightMap  = heightMap
    end,
    ComputeHeights = function(self)
        local plotNumbers = {}
        local totalPlots = self.iH * self.iW
        local min = math.huge
        local max = 0
        for plotIndex = 1, totalPlots do
            local height = self.heightMap[plotIndex]
            if not plotNumbers[height] then plotNumbers[height] = 1 end
            plotNumbers[height] = plotNumbers[height] + 1
            max = math.max(max, height)
            min = math.min(min, height)
        end
        local sea = 0
        local land = 0
        local hill = 0
        for i = min, max do
            sea = sea + (plotNumbers[i] or 0)
            if sea >= self.expectedSea then self.landPlotHeight = i break end
        end
        for i = self.landPlotHeight + 1, max do
            land = land + (plotNumbers[i] or 0)
            if land >= self.expectedLand then self.hillPlotHeight = i break end
        end
        for i = self.hillPlotHeight + 1, max do
            hill = hill + (plotNumbers[i] or 0)
            if hill >= self.expectedHill then self.mountainPlotHeight = i break end
        end
    end,
    Hotspots = function(self)
        local hotspots = self.plateNumber
        for i = 0, hotspots do
            local plotIndex = Map.Rand(self.iH, "y for hotspot") * self.iW + Map.Rand(self.iW, "x for hotspot")
            self.heightMap[plotIndex] = self.heightMap[plotIndex] + self.landLevel
        end
    end,
    ComputePlots = function(self)
        Wait( "L = "..tostring(self.landPlotHeight)..", H = "..tostring(self.hillPlotHeight)..", M = "..tostring(self.mountainPlotHeight) )
        self:ComputeHeights()
        Wait( "L = "..tostring(self.landPlotHeight)..", H = "..tostring(self.hillPlotHeight)..", M = "..tostring(self.mountainPlotHeight) )
        for y = 0, self.iH - 1 do
            for x = 0, self.iW - 1 do
                local plotIndex = y * self.iW + x + 1
                local height =  self.heightMap[plotIndex]
                if height < self.landPlotHeight then
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_OCEAN
                elseif height < self.hillPlotHeight then 
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_LAND
                elseif height < self.mountainPlotHeight then
                    self.plotTypes[plotIndex] = PlotTypes.PLOT_HILLS
                else
                    if Map.Rand(10, "random pass in mountain") > 7 then
                        self.plotTypes[plotIndex] = PlotTypes.PLOT_HILLS
                    else
                        self.plotTypes[plotIndex] = PlotTypes.PLOT_MOUNTAIN
                    end
                end
            end
        end
    end,
}
----------------------------------------------------------------------------------
function GeneratePlotTypes()
    print("Setting Plot Types (Tectonics) ...")

    local plotTypes = Voronoi:Create()
    SetPlotTypes(plotTypes)
    local args = { bExpandCoasts = false }
    GenerateCoasts(args) -- will have to look into that as I want to avoid coasts spanning into oceans as they currently do too easily

end
----------------------------------------------------------------------------------
function GenerateTerrain()
    print("Generating Terrain (Using default for the moment) ...")
    
    local terraingen = TerrainGenerator.Create()
    local terrainTypes = terraingen:GenerateTerrain()
        
    SetTerrainTypes(terrainTypes)
end
----------------------------------------------------------------------------------
function AddFeatures()
    print("Adding Features (using default implementation) ...")
    
    local featuregen = FeatureGenerator.Create()
    featuregen:AddFeatures()
end
